#include <rtl.h>
#include <rtl_fifo.h>
#include <rtl_sched.h>
#include <asm/io.h>


#define NUM_PWM 12
#define MAX_PWM 1000000
#define INVERT_MASK 0x0b00


struct pwm_t {
  hrtime_t length;
  int mask;
  hrtime_t lenbuf;
  int marked;
};

struct pwm_t pwm[NUM_PWM+1] = {
  {MAX_PWM,0x001,0,0},{MAX_PWM,0x002,0,0},
  {MAX_PWM,0x004,0,0},{MAX_PWM,0x008,0,0},
  {MAX_PWM,0x010,0,0},{MAX_PWM,0x020,0,0},
  {MAX_PWM,0x040,0,0},{MAX_PWM,0x080,0,0},
  {MAX_PWM,0x100,0,0},{MAX_PWM,0x200,0,0},
  {MAX_PWM,0x400,0,0},{MAX_PWM,0x800,0,0},
  {MAX_PWM,0,0,0}
};


pthread_t pwm_thread;
int port=0x378;
int fifo=0;


void *pwm_thread_code(void *t) {

  hrtime_t now=gethrtime();

  while(1) {

    int i,mask=0;
    hrtime_t t_old=0;

    for(i=0;i<NUM_PWM+1;i++) {
      pwm[i].lenbuf=pwm[i].length;
      pwm[i].marked=0;
    };
    
    for(i=0;i<NUM_PWM+1;i++) {
      int j,channel=NUM_PWM;
      hrtime_t t_new=MAX_PWM;
      for(j=0;j<NUM_PWM+1;j++) {
	if(pwm[j].marked==0) {
	  if(pwm[j].lenbuf<=t_new) {
	    t_new=pwm[j].lenbuf;
	    channel=j;
	  };
	};
      };
      if(t_new!=t_old) {
	int v=mask^INVERT_MASK;
	outb(v&0xff,port);
	outb(v>>8,port+2);
	pthread_make_periodic_np(pthread_self(),
				 now+=(t_new-t_old),MAX_PWM);
	pthread_wait_np();
	t_old=t_new;
      };
      mask|=pwm[channel].mask;
      pwm[channel].marked=1;
    };

  };

};


int pwm_fifo_handler(unsigned int fifo) {

  struct {unsigned char channel,data;} msg;
  int err;
  
  while ((err = rtf_get(fifo,&msg,sizeof(msg))) == sizeof(msg)) {
    if(msg.channel>=NUM_PWM) return -EINVAL;
    pwm[msg.channel].length=(MAX_PWM*((int)msg.data))/255;
  };
    
  if (err != 0) {
    return -EINVAL;
  }

  return 0;
}


MODULE_PARM(port,"i");
MODULE_PARM(fifo,"i");


int init_module(void) {

  pthread_attr_t attr;
  struct sched_param sched_param;

  rtf_create(fifo,4000);
  rtf_create_handler(fifo,&pwm_fifo_handler);

  pthread_attr_init(&attr);
  pthread_attr_setcpu_np(&attr,0);
  sched_param.sched_priority=1;
  pthread_attr_setschedparam(&attr,&sched_param);
  pthread_create(&pwm_thread,&attr,pwm_thread_code,0);

  return 0;
};


void cleanup_module(void) {
  pthread_delete_np(pwm_thread);
  rtf_destroy(fifo);
};


