#include <rtl.h>
#include <rtl_fifo.h>
#include <rtl_sched.h>
#include <asm/io.h>

#include "servodata.h"

struct servodata servo_data[NUM_SERVOS]={
  {ONE_MS,0x80,0x0b},{ONE_MS,0x40,0x0b},{ONE_MS,0x20,0x0b},
  {ONE_MS,0x10,0x0b},{ONE_MS,0x08,0x0b},{ONE_MS,0x00,0x03},
  {ONE_MS,0x04,0x0b},{ONE_MS,0x00,0x0f},{ONE_MS,0x02,0x0b},
  {ONE_MS,0x01,0x0b},{ONE_MS,0x00,0x0a},{ONE_MS,0x00,0x09}};


pthread_t servo_thread;
int port=0x378;
int fifo=0;


void *servo_thread_code(void *t) {

  hrtime_t now=gethrtime();
 
  while(1) {

    int servo;
    for(servo=0;servo<NUM_SERVOS;servo++) {
      
      outb(servo_data[servo].mask1,port);
      outb(servo_data[servo].mask2,port+2);
      pthread_make_periodic_np(pthread_self(),
			       now+=servo_data[servo].len,30*ONE_MS);
      pthread_wait_np();

    };
  };
};

int servo_fifo_handler(unsigned int fifo) {

  struct {unsigned char channel,data;} msg;
  int err;

  while ((err = rtf_get(fifo,&msg,sizeof(msg))) == sizeof(msg)) {
    if(msg.channel>=NUM_SERVOS) return -EINVAL;
    servo_data[msg.channel].len=ONE_MS+((ONE_MS*msg.data)/255);
  };
    
  if (err != 0) {
    return -EINVAL;
  }

  return 0;
}


MODULE_PARM(port,"i");
MODULE_PARM(fifo,"i");


int init_module(void) {

  pthread_attr_t attr;
  struct sched_param sched_param;

  rtf_create(fifo,4000);
  rtf_create_handler(fifo,&servo_fifo_handler);

  pthread_attr_init(&attr);
  pthread_attr_setcpu_np(&attr,0);
  sched_param.sched_priority = 1;
  pthread_attr_setschedparam(&attr, &sched_param);

  return pthread_create(&servo_thread,&attr,servo_thread_code,0);

};


void cleanup_module(void) {
  pthread_delete_np(servo_thread);
  rtf_destroy(fifo);
};










